#include<stdio.h>
#include<assert.h>
#include<string.h>

#include"nb_key.h"
#include"calibration.h"

#include <HD/hd.h>

#include <HDU/hduVector.h>
#include <HDU/hduError.h>

/* Struktura do przekazywania danych pobranych z urzadzenia. */
typedef struct 
{
    HDboolean m_buttonState;        /* Nacisniecie przycisku.	*/
    hduVector3Dd m_devicePosition;  /* Wspolrzedne.				*/
    hduVector3Dd m_deviceAngles;    /* Katy (gimbals).			*/
    HDdouble m_deviceTransform[16]; /* Macierz obrotu.			*/
    HDErrorInfo m_error;			/* Blad.					*/
} DeviceData;

/* Do tej struktury maja dostep tylko callbaki */
static DeviceData gServoDeviceData;

/* Callback aktualizujacy strukture gServoDeviceData */
HDCallbackCode HDCALLBACK updateDeviceCallback(void *pUserData)
{   
    int nButtons = 0;

	/* Poczatek ramki haptycznej */
    hdBeginFrame(hdGetCurrentDevice());

    hdGetIntegerv(HD_CURRENT_BUTTONS, &nButtons);
    
    /* Wyciagamy stan pierwszego przycisku */
    gServoDeviceData.m_buttonState = (nButtons & HD_DEVICE_BUTTON_1) ? HD_TRUE : HD_FALSE;
        
    hdGetDoublev(HD_CURRENT_POSITION, gServoDeviceData.m_devicePosition);
    hdGetDoublev(HD_CURRENT_GIMBAL_ANGLES, gServoDeviceData.m_deviceAngles);
    hdGetDoublev(HD_CURRENT_TRANSFORM, gServoDeviceData.m_deviceTransform);

    gServoDeviceData.m_error = hdGetError();

    /* Koniec ramki. */
    hdEndFrame(hdGetCurrentDevice());

    /* Ten callback ma byc wywolany rowniez w nastepnej iteracji schedulera. */
    return HD_CALLBACK_CONTINUE;    
}

/* Callback wykopiowuje gServoDeviceData. */
HDCallbackCode HDCALLBACK copyDeviceDataCallback(void *pUserData)
{
    DeviceData *pDeviceData = (DeviceData *) pUserData;

    memcpy(pDeviceData, &gServoDeviceData, sizeof(DeviceData));

    /* Ten callback jest jednorazowy. */
    return HD_CALLBACK_DONE;
}


int main(int argc, char* argv[])
{
    HDSchedulerHandle hUpdateHandle = 0;
    HDErrorInfo error;

    /* Inicjalizacja urzadzenia. Od tego zaczynamy */
    HHD hHD = hdInitDevice(HD_DEFAULT_DEVICE);
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to initialize haptic device");
        return -1;
    }

	/* Udalo sie. Urzadzenie jest i dziala. */
    printf("Hello Haptic Device!\n");
    printf("Found device model: %s.\n\n", hdGetString(HD_DEVICE_MODEL_TYPE));

	/* Start petli urzadzenia */
    hdEnable(HD_FORCE_OUTPUT);
    hdStartScheduler();

    /* Czy sie udalo? */
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to start scheduler");
        fprintf(stderr, "\nPress any key to quit.\n");
        return -1;
    }

	/* Kalibracja urzadzenia. Co bedzie, jesli sie ja wylaczy? */
	calibrate_haptic();

    /* Instalujemy callback. 0 oznacza pusty wskaznik do danych, ktore moglyby byc zwrocone. */
	hUpdateHandle = hdScheduleAsynchronous(updateDeviceCallback, 0, HD_MAX_SCHEDULER_PRIORITY);

    int kTerminateCount = 1000;
    int buttonHoldCount = 0;

    /* Dane pobrane z urzadzenia (biezace i poprzednie) */
    DeviceData currentData;
    DeviceData prevData;

    /* 
	 * Wykopiowujemy dane z gServoDeviceData do currentData. 
	 * To wywolanie jest synchroniczne. 
	 */
    hdScheduleSynchronous(copyDeviceDataCallback, &currentData, HD_MIN_SCHEDULER_PRIORITY);

    memcpy(&prevData, &currentData, sizeof(DeviceData));    

	printf("Press and release the stylus button to print out the current device location.\nPress and hold the stylus button to exit the application\n");

    /* Ta petla trwa dopoki uzytkownik nie przytrzyma przycisku wystarczajaco dlugo. */
    while (1) {
    	/* Wykopiowujemy dane z gServoDeviceData do currentData. */
        hdScheduleSynchronous(copyDeviceDataCallback, &currentData, HD_MIN_SCHEDULER_PRIORITY);

        /* Przycisk zostal puszczony. Wyswietlamy polozenie. */
        if(currentData.m_buttonState && !prevData.m_buttonState) {           
            printf("Current position: (%g, %g, %g)\n", 
                currentData.m_devicePosition[0], 
                currentData.m_devicePosition[1], 
                currentData.m_devicePosition[2]); 

            printf("Current angles: (%g, %g, %g)\n", 
                currentData.m_deviceAngles[0]/3.14*180, 
                currentData.m_deviceAngles[1]/3.14*180, 
                currentData.m_deviceAngles[2]/3.14*180); 

            printf("Current matrix: ");

			for(int i=0; i<16; i++) {
				if(!(i%4)) printf("\n");
            	printf("%g\t", currentData.m_deviceTransform[i]);
			}

			printf("\n");
        } else if (currentData.m_buttonState && prevData.m_buttonState) {
			/* Uzytkownik trzyma przycisk. Liczymy. */
            buttonHoldCount++;

            if (buttonHoldCount > kTerminateCount) {
                /* Przycisk byl trzymany wystarczajaco dlugo. Koniec. */
                break;
            }
        } else if (!currentData.m_buttonState && prevData.m_buttonState) {
            /* Przycisk zostal wcisniety. Zerujemy licznik. */
            buttonHoldCount = 0;
        }

		/* Sprawdzenie, czy nie wystapil blad. */
		if (HD_DEVICE_ERROR(currentData.m_error)) {
            hduPrintError(stderr, &currentData.m_error, "Device error detected");

			if (hduIsSchedulerError(&error)) {
                fprintf(stderr, "\nQuitting.\n");
                break;
			}
		}

        /* Przepisujemy biezace dane na poprzednie. */
        memcpy(&prevData, &currentData, sizeof(DeviceData));    
    }

    /* Koncowe sprzatanie. */
    hdStopScheduler();
    hdUnschedule(hUpdateHandle);
    hdDisableDevice(hHD);

    return 0;
}

